/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Frontend - infinite-login.service.ts
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

import {
  Inject,
  Injectable,
} from "@angular/core";
import {Navigate} from "@ngxs/router-plugin";
import {Store} from "@ngxs/store";
import {AppRoutesPartialEnum} from "../enums/partial/app-routes-partial.enum";
import {DefaultSolidifyEnvironment} from "../environments/environment.solidify-defaults";
import {UrlQueryParamHelper} from "../helpers/url-query-param.helper";
import {ENVIRONMENT} from "../injection-tokens/environment.injection-token";
import {AbstractBaseService} from "../services/abstract-base.service";
import {
  isNotNullNorUndefinedNorWhiteString,
  isNullOrUndefined,
  isNullOrUndefinedOrWhiteString,
} from "../tools/is/is.tool";
import {MappingObjectUtil} from "../utils/mapping-object.util";

@Injectable({
  providedIn: "root",
})
export class InfiniteLoginService extends AbstractBaseService {
  private readonly _QUERY_PARAM_OAUTH_REMAINING_TRY_LOGGING: string = "remaining_try_logging";

  private _currentRemainingTryLogging: number;

  constructor(@Inject(ENVIRONMENT) private readonly _environment: DefaultSolidifyEnvironment,
              private readonly _store: Store) {
    super();
  }

  generateStateWithRemainingTryLogging(state: string): string {
    if (this._isDisabledFeature) {
      return state;
    }
    let finalState = "";
    let separator = "?";
    if (isNotNullNorUndefinedNorWhiteString(state)) {
      finalState = state;
      if (state.includes("?")) {
        separator = "&";
      }
    }
    return finalState + `${separator}${this._QUERY_PARAM_OAUTH_REMAINING_TRY_LOGGING}=${this._currentRemainingTryLogging}`;
  }

  resetRemainingTryLogging(): void {
    if (this._isDisabledFeature) {
      return;
    }
    this._currentRemainingTryLogging = this._environment.initialValueRemainingTryLogging;
  }

  decrementTryLogging(): void {
    if (this._isDisabledFeature) {
      return;
    }
    const redirectPathMappingObject = UrlQueryParamHelper.getQueryParamMappingObjectFromUrl(window.location.pathname + window.location.search);
    const isOAuth2TryLogging = MappingObjectUtil.has(redirectPathMappingObject, this._QUERY_PARAM_OAUTH_REMAINING_TRY_LOGGING);
    if (isOAuth2TryLogging) {
      const currentRemainingTryLogging = +MappingObjectUtil.get(redirectPathMappingObject, this._QUERY_PARAM_OAUTH_REMAINING_TRY_LOGGING);
      this._currentRemainingTryLogging = currentRemainingTryLogging - 1;
    }
  }

  extractRemainingTryLogging(redirectPath: string): string {
    if (this._isDisabledFeature) {
      return redirectPath;
    }
    const redirectPathMappingObject = UrlQueryParamHelper.getQueryParamMappingObjectFromUrl(redirectPath);
    const isOAuth2TryLogging = MappingObjectUtil.has(redirectPathMappingObject, this._QUERY_PARAM_OAUTH_REMAINING_TRY_LOGGING);
    if (isOAuth2TryLogging) {
      this._currentRemainingTryLogging = +MappingObjectUtil.get(redirectPathMappingObject, this._QUERY_PARAM_OAUTH_REMAINING_TRY_LOGGING);
      MappingObjectUtil.delete(redirectPathMappingObject, this._QUERY_PARAM_OAUTH_REMAINING_TRY_LOGGING);
      redirectPath = UrlQueryParamHelper.removeParametersFromPath(redirectPath);
      if (isNullOrUndefinedOrWhiteString(redirectPath)) {
        redirectPath = AppRoutesPartialEnum.home;
      }
      redirectPath = redirectPath + UrlQueryParamHelper.getQueryParamStringFromQueryParamMappingObject(redirectPathMappingObject);
    }
    return redirectPath;
  }

  redirectToUnableToLoadApp(): boolean {
    if (this._currentRemainingTryLogging > 0) {
      return false;
    }
    this._store.dispatch(new Navigate([AppRoutesPartialEnum.unableToLoadApp]));
    return true;
  }

  private get _isDisabledFeature(): boolean {
    return isNullOrUndefined(this._environment.initialValueRemainingTryLogging);
  }
}
