/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Frontend - abstract-app.component.ts
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

import {DOCUMENT} from "@angular/common";
import {
  AfterViewInit,
  ChangeDetectorRef,
  Directive,
  ElementRef,
  HostListener,
  Inject,
  Injector,
  OnInit,
  Renderer2,
  Type,
  ViewChild,
} from "@angular/core";
import {MatDialog} from "@angular/material/dialog";
import {
  Router,
  RouterStateSnapshot,
} from "@angular/router";
import {TranslateService} from "@ngx-translate/core";
import {Navigate} from "@ngxs/router-plugin";
import {Store} from "@ngxs/store";
import {
  combineLatest,
  fromEvent,
  Observable,
} from "rxjs";
import {
  distinctUntilChanged,
  filter,
  map,
  take,
  tap,
} from "rxjs/operators";
import {AppConfigService} from "../../../config/app-config.service";
import {SOLIDIFY_CONSTANTS} from "../../../constants";
import {CookieConsentService} from "../../../cookie-consent/services/cookie-consent.service";
import {BannerColorEnum} from "../../../enums/banner-color.enum";
import {LanguagePartialEnum} from "../../../enums/partial/language-partial.enum";
import {LocalStoragePartialEnum} from "../../../enums/partial/local-storage-partial.enum";
import {ThemePartialEnum} from "../../../enums/partial/theme-partial.enum";
import {ChemicalMoleculeVisualizationEnum} from "../../../file-preview/enums/chemical-molecule-visualization.enum";
import {LocalStorageHelper} from "../../../helpers/local-storage.helper";
import {PollingHelper} from "../../../helpers/polling.helper";
import {SessionStorageHelper} from "../../../helpers/session-storage.helper";
import {User} from "../../../models/dto/user.model";
import {FrontendVersion} from "../../../models/frontend-version.model";
import {GlobalBanner} from "../../../models/global-banner.model";
import {StateModel} from "../../../models/stores/state.model";
import {AppStatusService} from "../../../services/app-status.service";
import {BreakpointService} from "../../../services/breakpoint.service";
import {GoogleAnalyticsService} from "../../../services/google-analytics.service";
import {LoggingService} from "../../../services/logging.service";
import {NotificationService} from "../../../services/notification.service";
import {ScrollService} from "../../../services/scroll.service";
import {SolidifySecurityService} from "../../../services/security.service";
import {SolidifyAppUserStateModel} from "../../../stores/abstract/app-user/app-user-state.model";
import {SolidifyAppUserState} from "../../../stores/abstract/app-user/app-user.state";
import {SolidifyAppStateModel} from "../../../stores/abstract/app/app-state.model";
import {SolidifyAppAction} from "../../../stores/abstract/app/app.action";
import {SolidifyAppState} from "../../../stores/abstract/app/app.state";
import {AppBannerState} from "../../../stores/banner/app-banner.state";
import {
  isFalse,
  isFalsy,
  isNotNullNorUndefined,
  isNullOrUndefined,
  isTrue,
  isTruthyObject,
  isUndefined,
} from "../../../tools/is/is.tool";
import {ExtendEnum} from "../../../types/extend-enum.type";
import {SolidifyObject} from "../../../types/solidify-object.type";
import {DateUtil} from "../../../utils/date.util";
import {DialogUtil} from "../../../utils/dialog.util";
import {MemoizedUtil} from "../../../utils/stores/memoized.util";
import {UpdateVersionDialog} from "../../dialogs/update-version/update-version.dialog";
import {AbstractInternalContainer} from "../abstract-internal/abstract-internal.container";

declare let gtag: (name: string, id: string, option: SolidifyObject) => void;

// @dynamic
@Directive()
export abstract class AbstractAppComponent extends AbstractInternalContainer implements OnInit, AfterViewInit {

  protected readonly _ATTRIBUTE_SRC_DIFFER_KEY: string = "src-differ";
  protected readonly _ATTRIBUTE_SRC: string = "src";

  private static readonly _THEME_ATTRIBUTE_NAME: string = "theme";
  private static readonly _DARK_MODE_ATTRIBUTE_NAME: string = "dark-mode";
  private static readonly _UI_ATTRIBUTE_NAME: string = "ui";
  logo: string;

  urlStateObs: Observable<RouterStateSnapshot> = this._store.select((state: StateModel) => state.router.state);
  appLanguageObs: Observable<ExtendEnum<LanguagePartialEnum>>;
  isLoggedObs: Observable<boolean>;
  isApplicationInitializedObs: Observable<boolean>;
  isServerOfflineObs: Observable<boolean>;
  isUnableToLoadAppObs: Observable<boolean>;
  themeObs: Observable<ExtendEnum<ThemePartialEnum>>;
  globalBannerObs: Observable<GlobalBanner>;
  darkModeObs: Observable<boolean>;
  currentUserObs: Observable<User>;
  isTouchInterfaceObs: Observable<boolean>;

  frontendVersionObs: Observable<FrontendVersion>;
  displayBannerObs: Observable<boolean> = MemoizedUtil.select(this._store, AppBannerState, state => state.display);
  messageBannerObs: Observable<string> = MemoizedUtil.select(this._store, AppBannerState, state => state.message);
  parametersObs: Observable<SolidifyObject> = MemoizedUtil.select(this._store, AppBannerState, state => state.parameters);
  colorBannerObs: Observable<BannerColorEnum> = MemoizedUtil.select(this._store, AppBannerState, state => state.color);

  ignoreGrid: boolean = false;
  displayCart: boolean = false;
  isMaintenanceModeActive: boolean = false;

  @ViewChild("main")
  mainElementRef: ElementRef;

  @HostListener("window:beforeunload", ["$event"])
  leaveWithoutSaveNotification($event: Event): void {
    const preventExit = MemoizedUtil.selectSnapshot(this._store, this.environment.appState, state => state.preventExit);
    const ignorePreventExit = MemoizedUtil.selectSnapshot(this._store, this.environment.appState, state => state.ignorePreventLeavePopup);
    if (isFalse(ignorePreventExit) && isTrue(preventExit)) {
      $event.returnValue = true;
    }
  }

  get sessionStorageHelper(): typeof SessionStorageHelper {
    return SessionStorageHelper;
  }

  get localStorageHelper(): typeof LocalStorageHelper {
    return LocalStorageHelper;
  }

  constructor(protected readonly _appState: Type<SolidifyAppState<SolidifyAppStateModel>>,
              protected readonly _appUserState: Type<SolidifyAppUserState<SolidifyAppUserStateModel>>,
              protected readonly _store: Store,
              protected readonly _router: Router,
              protected readonly _translate: TranslateService,
              protected readonly _renderer: Renderer2,
              @Inject(DOCUMENT) protected readonly _document: Document,
              protected readonly _injector: Injector,
              protected readonly _notificationService: NotificationService,
              protected readonly _loggingService: LoggingService,
              public readonly appStatusService: AppStatusService,
              public readonly breakpointService: BreakpointService,
              protected readonly _dialog: MatDialog,
              protected readonly _scrollService: ScrollService,
              protected readonly _googleAnalyticsService: GoogleAnalyticsService,
              protected readonly _cookieConsentService: CookieConsentService,
              protected readonly _appConfigService: AppConfigService,
              protected readonly _changeDetector: ChangeDetectorRef,
              protected readonly _securityService: SolidifySecurityService,
  ) {
    super(_injector);
    new DateUtil(); // Allow to initiate moment in DateUtil
    this.environment.appState = this._appState;
    this.environment.appUserState = this._appUserState;
    this.frontendVersionObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.frontendVersion);
    this.appLanguageObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.appLanguage);
    this.isLoggedObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.isLoggedIn);
    this.isApplicationInitializedObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.isApplicationInitialized);
    this.isServerOfflineObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.isServerOffline);
    this.isUnableToLoadAppObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.isUnableToLoadApp);
    this.themeObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.theme);
    this.globalBannerObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.globalBanner);
    this.darkModeObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.darkMode);
    if (isNotNullNorUndefined(this.environment.appUserState)) {
      this.currentUserObs = MemoizedUtil.current(this._store, this.environment.appUserState);
    }
    this.isTouchInterfaceObs = MemoizedUtil.select(this._store, this.environment.appState, state => state.isTouchInterface);
    this._store.dispatch(new SolidifyAppAction.InitApplication());
    this._setOverrideCssStyleSheet();
    this.subscribe(this._observeThemeChange());
    this.subscribe(this._observeDarkModeChange());
    this.observeOfflineOnlineMode();
    this.themeChange(this.environment.theme);
    this._initDarkMode();
    this.subscribe(this.observeUpdateAvailableForModal());
    this.subscribe(this._observeGlobalBanner());
    this.subscribe(this.listenLanguageChange());
    this.subscribe(this.listenAppReady());
    this._cookieConsentService.openBannerIfNeverAnswer();
    this._googleAnalyticsService.init(this._renderer);
    this._activeChemicalMoleculePreviewIfEnable();
  }

  ngOnInit(): void {
    super.ngOnInit();

    this.subscribe(this._observeToUpdateMaintenanceMode());
    this.subscribe(this.isTouchInterfaceObs, isTouchInterface => this.setTouchInterface(isTouchInterface));
    this.subscribe(this.observeTouchEvent());
  }

  private _observeToUpdateMaintenanceMode(): Observable<any> {
    this.isMaintenanceModeActive = isTrue(this.environment.maintenanceMode);
    return combineLatest([this.isApplicationInitializedObs, this._appConfigService.environmentUpdatedObs])
      .pipe(
        map(([isInitialized, environment]) => {
          this.isMaintenanceModeActive = isTrue(this.environment.maintenanceMode);
          this._changeDetector.detectChanges();
          return this.isMaintenanceModeActive;
        }),
      );
  }

  ngAfterViewInit(): void {
    super.ngAfterViewInit();
    if (isTruthyObject(this.mainElementRef)) {
      this._scrollService.init(this.mainElementRef);
    }
  }

  private observeTouchEvent(): Observable<Event> {
    return fromEvent(document, "touchstart")
      .pipe(
        take(1),
        tap(event => this._store.dispatch(new SolidifyAppAction.SetIsTouchInterface(true)),
        ),
      );
  }

  private observeOfflineOnlineMode(): void {
    let previousStateIsOnline: boolean | undefined = undefined;
    this.subscribe(this.appStatusService.online.pipe(
      distinctUntilChanged(),
      tap((isOnline: boolean) => {
        if ((isUndefined(previousStateIsOnline) || isTrue(previousStateIsOnline)) && isFalsy(isOnline)) {
          this._notificationService.showWarning(this.labelTranslateInterface.appOffline);
          previousStateIsOnline = isOnline;
          return;
        }

        if (isFalse(previousStateIsOnline) && isTrue(isOnline)) {
          this._notificationService.showInformation(this.labelTranslateInterface.appBackToOnline);
          previousStateIsOnline = isOnline;
          return;
        }
        return;
      }),
    ));
  }

  private _observeGlobalBanner(): Observable<any> {
    return PollingHelper.startPollingObs({
      initialIntervalRefreshInSecond: this.environment.refreshGlobalBannerIntervalInSecond,
      incrementInterval: true,
      resetIntervalWhenUserMouseEvent: true,
      maximumIntervalRefreshInSecond: this.environment.refreshGlobalBannerIntervalInSecond * 10,
      filter: () => MemoizedUtil.selectSnapshot(this._store, this.environment.appState, state => isFalse(state.isServerOffline))
        && (this._securityService.isRoot() || isFalse(this.environment.maintenanceMode)),
      actionToDo: () => {
        this._store.dispatch(new SolidifyAppAction.GetGlobalBanner());
      },
    });
  }

  private observeUpdateAvailableForModal(): Observable<string> {
    return this.appStatusService.updateAvailable.pipe(
      distinctUntilChanged(),
      filter(version => !isNullOrUndefined(version)),
      tap((version) => {
        switch (this.environment.actionToDoWhenUpdateAvailable) {
          case "do-nothing":
            break;
          case "prompt":
            DialogUtil.open(this._dialog, UpdateVersionDialog);
            break;
          case "refresh":
          default:
            window.location.reload();
            break;
        }
      }),
    );
  }

  useLanguage(language: ExtendEnum<LanguagePartialEnum>): void {
    this._store.dispatch(new SolidifyAppAction.ChangeAppLanguage(language));
  }

  navigate(route: string): void {
    this._store.dispatch(new Navigate([route]));
  }

  private _setTitle(): void {
    setTimeout(() => {
      this._document.title = this.environment.appTitle;
      const description = this.environment.appDescription;
      this._document.querySelector("meta[name=\"description\"]")?.setAttribute("content", description);
    }, 0);
  }

  private setTheme(): void {
    this._renderer.setAttribute(this._document.body, AbstractAppComponent._THEME_ATTRIBUTE_NAME, String(this.environment.theme));
  }

  private setDarkMode(enabled: boolean): void {
    this._renderer.setAttribute(this._document.body, AbstractAppComponent._DARK_MODE_ATTRIBUTE_NAME, enabled ? "true" : "false");
  }

  private setTouchInterface(isTouchInterface: boolean): void {
    this._renderer.setAttribute(this._document.body, AbstractAppComponent._UI_ATTRIBUTE_NAME, isTouchInterface ? "touch" : "standard");
  }

  private setFavicon(): void {
    const header = this._document.querySelector("head");
    const oldFavicon = this._document.querySelector("head link[type='image/x-abstract-icon']");
    let link = this._document.createElement("link");
    if (isNotNullNorUndefined(oldFavicon)) {
      link = oldFavicon as any;
    }
    this._renderer.setAttribute(link, "type", "image/x-abstract-icon");
    this._renderer.setAttribute(link, "rel", "shortcut abstract-icon");
    this._renderer.setAttribute(link, "href", `./assets/themes/${this.environment.theme}/favicon.ico`);
    this._renderer.appendChild(header, link);
  }

  private setImageToolbar(): void {
    this.logo = `assets/themes/${this.environment.theme}/toolbar-header-image.png`;
  }

  logout(): void {
    this._store.dispatch(new SolidifyAppAction.Logout());
  }

  getTitle(): string {
    return this._document.title;
  }

  darkThemeChange(enabled: boolean): void {
    this._store.dispatch(new SolidifyAppAction.ChangeDarkMode(enabled));
  }

  themeChange(theme: ExtendEnum<ThemePartialEnum>): void {
    if (theme !== MemoizedUtil.selectSnapshot(this._store, this.environment.appState, state => state.theme)) {
      this._store.dispatch(new SolidifyAppAction.ChangeAppTheme(theme));
    }
  }

  navigateToHome(): void {
    this.navigate(this.environment.routeHomePage);
  }

  private _observeThemeChange(): Observable<ExtendEnum<ThemePartialEnum>> {
    return this.themeObs.pipe(
      distinctUntilChanged(),
      tap((theme: ExtendEnum<ThemePartialEnum>) => {
        this.setTheme();
        this.setFavicon();
        this.setImageToolbar();
      }),
    );
  }

  private _observeDarkModeChange(): Observable<boolean> {
    return this.darkModeObs.pipe(
      distinctUntilChanged(),
      tap((enabled: boolean) => {
        this.setDarkMode(enabled);
      }),
    );
  }

  private listenLanguageChange(): Observable<ExtendEnum<LanguagePartialEnum>> {
    return this.appLanguageObs.pipe(
      tap((language: ExtendEnum<LanguagePartialEnum>) => {
        this._setTitle();
      }),
    );
  }

  private listenAppReady(): Observable<boolean> {
    return this.isApplicationInitializedObs.pipe(
      tap((isInit: boolean) => {
        this._setTitle();
      }),
    );
  }

  private _setOverrideCssStyleSheet(): void {
    const header = this._document.querySelector("head");
    const link = this._document.createElement("link");
    this._renderer.setAttribute(link, "type", "text/css");
    this._renderer.setAttribute(link, "rel", "stylesheet");
    this._renderer.setAttribute(link, "href", `./assets/styles/override.css`);
    this._renderer.appendChild(header, link);
  }

  protected _getValueScriptDiffer(script: Element): string | undefined {
    if (isNullOrUndefined(script) || isNullOrUndefined(script.attributes)) {
      return undefined;
    }
    const attribute = script.attributes[this._ATTRIBUTE_SRC_DIFFER_KEY];
    return isNullOrUndefined(attribute) ? undefined : attribute.value;
  }

  private _activeChemicalMoleculePreviewIfEnable(): void {
    if (this.environment.visualizationChemicalMoleculeMode === ChemicalMoleculeVisualizationEnum.disabled) {
      return;
    }
    const threeDimensionalLib = this.environment.visualizationChemicalMolecule3dLib;
    const twoDimensionalLibs = this.environment.visualizationChemicalMolecule2dLibs;
    const listScript = [];
    this._document.querySelectorAll(`script[${this._ATTRIBUTE_SRC_DIFFER_KEY}]`).forEach(script => {
      listScript.push(script);
    });

    if (this.environment.visualizationChemicalMoleculeMode === ChemicalMoleculeVisualizationEnum.threeDimensionalOnly
      || this.environment.visualizationChemicalMoleculeMode === ChemicalMoleculeVisualizationEnum.threeAndTwoDimensional) {
      const jsmolScript = this._getScriptWithValueEndWith(listScript, threeDimensionalLib);
      if (isNullOrUndefined(jsmolScript)) {
        this._loggingService.logWarning(`Unable to enable chemical molecule visualization, missing ${threeDimensionalLib}`);
        return;
      }
      this._activeScriptDiffer([jsmolScript]);
    }

    if (this.environment.visualizationChemicalMoleculeMode === ChemicalMoleculeVisualizationEnum.threeAndTwoDimensional) {
      const listScriptToTreat = [];
      twoDimensionalLibs.forEach(libNameNeeded => {
        const scriptLibNeeded = this._getScriptWithValueEndWith(listScript, libNameNeeded);
        if (isNullOrUndefined(scriptLibNeeded)) {
          this._loggingService.logWarning(`Unable to enable chemical molecule visualization 2D, missing ${libNameNeeded}`);
          return;
        }
        listScriptToTreat.push(scriptLibNeeded);
      });
      if (listScriptToTreat.length !== twoDimensionalLibs.length) {
        this._loggingService.logWarning(`Swith to chemical molecule visualization 3D`);
        this.environment.visualizationChemicalMoleculeMode = ChemicalMoleculeVisualizationEnum.threeDimensionalOnly;
        return;
      }
      // Need to differ lib JSME to avoid error at startup
      setTimeout(() => {
        this._activeScriptDiffer(listScriptToTreat);
      }, 1000);
    }
  }

  private _activeScriptDiffer(listScript: Element[]): void {
    listScript.forEach(script => {
      const value = this._getValueScriptDiffer(script);
      if (isNullOrUndefined(value)) {
        return;
      }
      this._renderer.setAttribute(script, this._ATTRIBUTE_SRC, value);
      this._renderer.removeAttribute(script, this._ATTRIBUTE_SRC_DIFFER_KEY);
    });
  }

  protected _getScriptWithValueEndWith(listScript: Element[], threeDimensionalLib: string): Element | undefined {
    return listScript.find(script => {
      const value = this._getValueScriptDiffer(script);
      return isNotNullNorUndefined(value) && value.endsWith(threeDimensionalLib);
    });
  }

  scrollHandler($event: Event): void {
    this._scrollService.triggerScrollEventMain($event);
  }

  private _initDarkMode(): void {
    const darkModeInStorage = LocalStorageHelper.getItem(LocalStoragePartialEnum.darkMode);
    let darkMode = this.environment.darkMode;
    if (darkModeInStorage === SOLIDIFY_CONSTANTS.STRING_TRUE) {
      darkMode = true;
    } else if (darkModeInStorage === SOLIDIFY_CONSTANTS.STRING_FALSE) {
      darkMode = false;
    }
    this.darkThemeChange(darkMode);
  }
}
