/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Frontend - about.routable.ts
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

import {
  ChangeDetectionStrategy,
  Component,
  Inject,
  Injector,
  OnDestroy,
  OnInit,
} from "@angular/core";
import {Navigate} from "@ngxs/router-plugin";
import {
  Actions,
  Store,
} from "@ngxs/store";
import {Observable} from "rxjs";
import {
  filter,
  take,
  tap,
} from "rxjs/operators";
import {AppRoutesPartialEnum} from "../../../enums/partial/app-routes-partial.enum";
import {DefaultSolidifyEnvironment} from "../../../environments/environment.solidify-defaults";
import {ENVIRONMENT} from "../../../injection-tokens/environment.injection-token";
import {LABEL_TRANSLATE} from "../../../injection-tokens/label-to-translate.injection-token";
import {BackendModuleVersion} from "../../../models/backend-modules/backend-module-version.model";
import {FrontendVersion} from "../../../models/frontend-version.model";
import {LabelTranslateInterface} from "../../../models/label-translate.model";
import {NotificationService} from "../../../services/notification.service";
import {SolidifyAppAction} from "../../../stores/abstract/app/app.action";
import {
  isNullOrUndefinedOrWhiteString,
  isTrue,
} from "../../../tools/is/is.tool";
import {MappingObject} from "../../../types/mapping-type.type";
import {ClipboardUtil} from "../../../utils/clipboard.util";
import {MappingObjectUtil} from "../../../utils/mapping-object.util";
import {MemoizedUtil} from "../../../utils/stores/memoized.util";
import {ofSolidifyActionCompleted} from "../../../utils/stores/store.tool";
import {StoreUtil} from "../../../utils/stores/store.util";
import {StringUtil} from "../../../utils/string.util";
import {AbstractInternalRoutable} from "../abstract-internal/abstract-internal.routable";

@Component({
  selector: "solidify-about-routable",
  templateUrl: "./about.routable.html",
  styleUrls: ["./about.routable.scss"],
  changeDetection: ChangeDetectionStrategy.OnPush,
})
export class AboutRoutable extends AbstractInternalRoutable implements OnInit, OnDestroy {
  backendVersionObs: Observable<MappingObject<string, BackendModuleVersion>> = MemoizedUtil.select(this._store, this._environment.appState, state => state.backendModulesVersion);
  frontendVersionObs: Observable<FrontendVersion> = MemoizedUtil.select(this._store, this._environment.appState, state => state.frontendVersion);

  getBackendVersionKey(backendVersion: MappingObject<string, BackendModuleVersion>): string[] {
    return MappingObjectUtil.keys(backendVersion);
  }

  getBackendVersionValues(backendVersion: MappingObject<string, BackendModuleVersion>): BackendModuleVersion[] {
    return MappingObjectUtil.values(backendVersion);
  }

  getBackendModuleVersion(moduleName: string, backendVersion: MappingObject<string, BackendModuleVersion>): BackendModuleVersion {
    return MappingObjectUtil.get(backendVersion, moduleName);
  }

  get currentTheme(): string {
    return StringUtil.convertToPascalCase(this._environment.theme);
  }

  get currentThemeName(): string {
    return isNullOrUndefinedOrWhiteString(this._environment.themeName) ? this.currentTheme : this._environment.themeName;
  }

  logo: string = `assets/themes/${this.environment.theme}/toolbar-header-image.png`;

  constructor(protected readonly _injector: Injector,
              @Inject(ENVIRONMENT) private readonly _environment: DefaultSolidifyEnvironment,
              @Inject(LABEL_TRANSLATE) readonly labelTranslateInterface: LabelTranslateInterface,
              private readonly notificationService: NotificationService,
              private readonly _store: Store,
              private readonly _actions$: Actions) {
    super(_injector);
  }

  ngOnInit(): void {
    super.ngOnInit();

    this.subscribe(StoreUtil.dispatchParallelActionAndWaitForSubActionsCompletion(this._store, [
      {
        action: new SolidifyAppAction.GetAppVersion(),
        subActionCompletions: [
          this._actions$.pipe(ofSolidifyActionCompleted(SolidifyAppAction.GetAppVersionSuccess)),
          this._actions$.pipe(ofSolidifyActionCompleted(SolidifyAppAction.GetAppVersionFail)),
        ],
      },
    ]));

    if (isTrue(this._environment.loadBackendModulesInfo)) {
      this.subscribe(MemoizedUtil.select(this._store, this._environment.appState, state => state.isApplicationInitialized).pipe(
        filter(isApplicationInitialized => isApplicationInitialized),
        take(1),
        tap(() => {
          const backendModulesUrl = MemoizedUtil.selectSnapshot(this._store, this._environment.appState, state => state.backendModulesUrl);
          this._store.dispatch(new SolidifyAppAction.LoadBackendModulesVersion(backendModulesUrl));
        }),
      ));
    }
  }

  copy(url: string): void {
    if (ClipboardUtil.copyStringToClipboard(url)) {
      this.notificationService.showInformation(this.labelTranslateInterface.notificationCopiedToClipboard);
    }
  }

  navigate(url: string[]): void {
    this._store.dispatch(new Navigate(url));
  }

  navigateToReleaseNotes(): void {
    this.navigate([AppRoutesPartialEnum.releaseNotes]);
  }

  navigateToChangelog(): void {
    this.navigate([AppRoutesPartialEnum.changelog]);
  }
}
