/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Frontend - abstract-create.routable.ts
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

import {
  ChangeDetectorRef,
  Directive,
  Injector,
  OnDestroy,
  OnInit,
} from "@angular/core";
import {Navigate} from "@ngxs/router-plugin";
import {
  Actions,
  Store,
} from "@ngxs/store";
import {Observable} from "rxjs";
import {StatePartialEnum} from "../../../enums/partial/state-partial.enum";
import {CrudHelper} from "../../../helpers/crud.helper";
import {UrlQueryParamHelper} from "../../../helpers/url-query-param.helper";
import {STORE_ROUTE_SERVICE} from "../../../injection-tokens/store-route-service.injection-token";
import {BaseResourceType} from "../../../models/dto/base-resource.model";
import {FormControlKey} from "../../../models/forms/form-control-key.model";
import {ModelFormControlEvent} from "../../../models/forms/model-form-control-event.model";
import {QueryParameters} from "../../../models/query-parameters/query-parameters.model";
import {AbstractStoreRouteService} from "../../../services/abstract-store-route-service";
import {ResourceActionHelper} from "../../../stores/abstract/resource/resource-action.helper";
import {ResourceNameSpace} from "../../../stores/abstract/resource/resource-namespace.model";
import {ResourceStateModel} from "../../../stores/abstract/resource/resource-state.model";
import {isNullOrUndefined} from "../../../tools/is/is.tool";
import {ExtendEnum} from "../../../types/extend-enum.type";
import {MappingObject} from "../../../types/mapping-type.type";
import {StoreUtil} from "../../../utils/stores/store.util";
import {AbstractCrudRoutable} from "../abstract-crud/abstract-crud.routable";

@Directive()
// eslint-disable-next-line @angular-eslint/directive-class-suffix
export abstract class AbstractCreateRoutable<TResourceModel extends BaseResourceType, UResourceStateModel extends ResourceStateModel<TResourceModel>>
  extends AbstractCrudRoutable<TResourceModel, UResourceStateModel> implements OnInit, OnDestroy {

  protected readonly _storeRouteService: AbstractStoreRouteService;

  isLoadingObs: Observable<boolean>;
  urlQueryParameters: MappingObject<string, string | undefined>;
  checkAvailableResourceNameSpace: ResourceNameSpace | undefined;
  alreadyUsedLabelToTranslate: string = this.labelTranslateInterface.alreadyUsed;

  protected constructor(protected readonly _store: Store,
                        protected readonly _actions$: Actions,
                        protected readonly _changeDetector: ChangeDetectorRef,
                        protected readonly _state: ExtendEnum<StatePartialEnum>,
                        protected readonly _injector: Injector,
                        private readonly _resourceActionNameSpace: ResourceNameSpace,
                        protected readonly _parentState?: ExtendEnum<StatePartialEnum>) {
    super(_store, _state, _injector, _parentState);
    this._storeRouteService = _injector.get(STORE_ROUTE_SERVICE);
    this.isLoadingObs = this._store.select(s => StoreUtil.isLoadingState(super.getState(s)));
  }

  create(modelFormControlEvent: ModelFormControlEvent<TResourceModel>): void {
    super.saveInProgress();
    this._store.dispatch(ResourceActionHelper.create(this._resourceActionNameSpace, modelFormControlEvent));
  }

  ngOnInit(): void {
    super.ngOnInit();
    this._store.dispatch(ResourceActionHelper.loadResource(this._resourceActionNameSpace));
    this._extractQueryParams();
  }

  ngOnDestroy(): void {
    super.ngOnDestroy();
    this._cleanState();
  }

  protected _cleanState(): void {
    this._store.dispatch(ResourceActionHelper.clean(this._resourceActionNameSpace, false, new QueryParameters()));
  }

  checkAvailable(formControlKey: FormControlKey): void {
    const resourceNameSpace = isNullOrUndefined(this.checkAvailableResourceNameSpace) ? this._resourceActionNameSpace : this.checkAvailableResourceNameSpace;
    this.subscribe(CrudHelper.checkAvailable(formControlKey, this._store, this._actions$, resourceNameSpace, undefined, this.alreadyUsedLabelToTranslate));
  }

  backToList(): void {
    this._store.dispatch(new Navigate([this._storeRouteService.getRootRoute(this._state)]));
  }

  protected _extractQueryParams(): void {
    this.urlQueryParameters = UrlQueryParamHelper.getQueryParamMappingObjectFromCurrentUrl();
  }
}
