/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Frontend - cookie-consent.service.ts
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

import {
  Inject,
  Injectable,
} from "@angular/core";
import {TranslateService} from "@ngx-translate/core";
import {Store} from "@ngxs/store";
import {CookieService} from "ngx-cookie";
import {
  BehaviorSubject,
  Observable,
} from "rxjs";
import {filter} from "rxjs/operators";
import {CookiePartialEnum} from "../../enums/partial/cookie-partial.enum";
import {LocalStoragePartialEnum} from "../../enums/partial/local-storage-partial.enum";
import {DefaultSolidifyEnvironment} from "../../environments/environment.solidify-defaults";
import {LocalStorageHelper} from "../../helpers/local-storage.helper";
import {SessionStorageHelper} from "../../helpers/session-storage.helper";
import {COOKIE_CONSENT_PREFERENCES} from "../../injection-tokens/cookie-consent-preferences.injection-token";
import {ENVIRONMENT} from "../../injection-tokens/environment.injection-token";
import {LABEL_TRANSLATE} from "../../injection-tokens/label-to-translate.injection-token";
import {LabelTranslateInterface} from "../../models/label-translate.model";
import {AbstractBaseService} from "../../services/abstract-base.service";
import {NotificationService} from "../../services/notification.service";
import {SolidifyAppAction} from "../../stores/abstract/app/app.action";
import {
  isFalse,
  isNotNullNorUndefinedNorWhiteString,
  isNullOrUndefined,
  isTrue,
} from "../../tools/is/is.tool";
import {ObservableUtil} from "../../utils/observable.util";
import {CookieType} from "../enums/cookie-type.enum";
import {CookiePreference} from "../models/cookie-preference.model";
import {CookieConsentUtil} from "../utils/cookie-consent.util";

@Injectable({
  providedIn: "root",
})
export class CookieConsentService extends AbstractBaseService {
  private readonly _cookiePreferenceUpdatedBS: BehaviorSubject<CookiePreference> = new BehaviorSubject<CookiePreference>(undefined);
  readonly cookiePreferenceUpdatedObs: Observable<CookiePreference> = ObservableUtil.asObservable(this._cookiePreferenceUpdatedBS);

  cookiePreferencePersonalizeCookie: CookiePreference = {
    type: CookieType.localStorage,
    isEnabled: false,
    name: LocalStoragePartialEnum.personalizeCookie,
    labelToTranslate: this.labelTranslate.cookieConsentSidebarPersonalizeCookieTitle,
    descriptionToTranslate: this.labelTranslate.cookieConsentSidebarPersonalizeCookieDescription,
    visible: () => true,
  };

  cookiePreferenceGoogleAnalytics: CookiePreference = {
    type: CookieType.cookie,
    isEnabled: false,
    name: CookiePartialEnum.googleAnalytics,
    labelToTranslate: this.labelTranslate.cookieConsentSidebarGoogleAnalyticsTitle,
    descriptionToTranslate: this.labelTranslate.cookieConsentSidebarGoogleAnalyticsDescription,
    visible: () => isNotNullNorUndefinedNorWhiteString(this._environment.googleAnalyticsId),
  };

  cookieDownloadToken: CookiePreference = {
    type: CookieType.cookie,
    isEnabled: false,
    name: CookiePartialEnum.downloadToken,
    labelToTranslate: this.labelTranslate.cookieConsentDownloadTokenTitle,
    descriptionToTranslate: this.labelTranslate.cookieConsentDownloadTokenDescription,
    visible: () => this._environment.useDownloadToken,
  };

  cookiePreferenceLanguage: CookiePreference = {
    type: CookieType.localStorage,
    isEnabled: false,
    name: LocalStoragePartialEnum.language,
    labelToTranslate: this.labelTranslate.cookieConsentSidebarLanguageTitle,
    descriptionToTranslate: this.labelTranslate.cookieConsentSidebarLanguageDescription,
    visible: () => true,
  };

  cookiePreferenceDarkMode: CookiePreference = {
    type: CookieType.localStorage,
    isEnabled: false,
    name: LocalStoragePartialEnum.darkMode,
    labelToTranslate: this.labelTranslate.cookieConsentSidebarDarkModeTitle,
    descriptionToTranslate: this.labelTranslate.cookieConsentSidebarDarkModeDescription,
    visible: () => true,
  };

  cookiePreferenceGlobalBanner: CookiePreference = {
    type: CookieType.localStorage,
    isEnabled: false,
    name: LocalStoragePartialEnum.ignoreGlobalBanner,
    labelToTranslate: this.labelTranslate.ignoreGlobalBannerTitle,
    descriptionToTranslate: this.labelTranslate.ignoreGlobalBannerDescription,
    visible: () => true,
  };

  cookiePreferencePrivacyPolicyAndTermsOfUseAccepted: CookiePreference = {
    type: CookieType.localStorage,
    isEnabled: false,
    name: LocalStoragePartialEnum.privacyPolicyAndTermsOfUseAccepted,
    labelToTranslate: this.labelTranslate.cookieConsentSidebarPrivacyPolicyAndTermsOfUseAcceptedTitle,
    descriptionToTranslate: this.labelTranslate.cookieConsentSidebarPrivacyPolicyAndTermsOfUseAcceptedDescription,
    visible: () => isTrue(this._environment.displayPrivacyPolicyAndTermsOfUseApprovalDialog),
  };

  listCookiePreferences: CookiePreference[] = [
    this.cookiePreferencePersonalizeCookie,
    this.cookieDownloadToken,
    this.cookiePreferenceGoogleAnalytics,
    this.cookiePreferenceLanguage,
    this.cookiePreferenceDarkMode,
    this.cookiePreferenceGlobalBanner,
    this.cookiePreferencePrivacyPolicyAndTermsOfUseAccepted,
    ...this._listCookiePreferencesInjected,
  ];

  constructor(private readonly _store: Store,
              private readonly _notificationService: NotificationService,
              private readonly _cookieService: CookieService,
              private readonly _translate: TranslateService,
              @Inject(LABEL_TRANSLATE) readonly labelTranslate: LabelTranslateInterface,
              @Inject(ENVIRONMENT) private readonly _environment: DefaultSolidifyEnvironment,
              @Inject(COOKIE_CONSENT_PREFERENCES) private readonly _listCookiePreferencesInjected: CookiePreference[],
  ) {
    super();
    this.initListCookiePreferencesFromLocalStorage();
  }

  get numberCookieEnabled(): number {
    return this.listCookiePreferences.filter(p => isTrue(p.isEnabled)).length;
  }

  get isAllCookiePreferencesAccepted(): boolean {
    return this.listCookiePreferences.findIndex(p => isFalse(p.isEnabled)) === -1;
  }

  watchCookiePreferenceUpdateObs(type: CookieType, name: string): Observable<CookiePreference> {
    return this.cookiePreferenceUpdatedObs.pipe(filter(p => p.type === type && p.name === name));
  }

  watchCookiePreferenceUpdateFromIdObs(id: string): Observable<CookiePreference> {
    return this.cookiePreferenceUpdatedObs.pipe(filter(p => CookieConsentUtil.generateCookiePreferenceId(p.type, p.name) === id));
  }

  isFeatureAuthorized(type: CookieType, name: string): boolean {
    return this.listCookiePreferences.findIndex(p => p.type === type && p.name === name && isTrue(p.isEnabled)) !== -1;
  }

  changeCookieSetting(cookiePreference: CookiePreference, isEnabled: boolean): void {
    const key = CookieConsentUtil.generateCookiePreferenceIdFromCookiePreference(cookiePreference);

    if (isTrue(isEnabled)) {
      cookiePreference.isEnabled = true;
      LocalStorageHelper.addItemInList(LocalStoragePartialEnum.personalizeCookie, key);
      this._enablePersonalizeCookieIfNot();
    } else {
      cookiePreference.isEnabled = false;
      LocalStorageHelper.removeItemInList(LocalStoragePartialEnum.personalizeCookie, key);

      if (cookiePreference.type === CookieType.cookie) {
        this._cookieService.remove(cookiePreference.name);
      } else if (cookiePreference.type === CookieType.localStorage) {
        LocalStorageHelper.removeItem(cookiePreference.name);
      } else if (cookiePreference.type === CookieType.sessionStorage) {
        SessionStorageHelper.removeItem(cookiePreference.name);
      }
    }
    this._cookiePreferenceUpdatedBS.next(cookiePreference);
  }

  private _enablePersonalizeCookieIfNot(): void {
    if (isFalse(this.cookiePreferencePersonalizeCookie.isEnabled)) {
      this.changeCookieSetting(this.cookiePreferencePersonalizeCookie, true);
    }
  }

  initListCookiePreferencesFromLocalStorage(): void {
    this.listCookiePreferences.forEach(p => {
      p.isEnabled = CookieConsentUtil.isFeatureAuthorizedFromCookiePreference(p);
    });
  }

  acceptAll(): void {
    this.listCookiePreferences.forEach(p => {
      this.changeCookieSetting(p, true);
    });
    this.hideBanner();
  }

  declineAll(): void {
    this.listCookiePreferences.forEach(p => {
      this.changeCookieSetting(p, false);
    });
    this.hideBanner();
  }

  openPersonalizationSidebar(): void {
    if (this.numberCookieEnabled === 0) {
      this.acceptAll();
    }
    this._store.dispatch(new SolidifyAppAction.ChangeDisplaySidebarCookieConsent(true));
  }

  closePersonalizationPanel(): void {
    this._store.dispatch(new SolidifyAppAction.ChangeDisplaySidebarCookieConsent(false));
  }

  openBannerIfNeverAnswer(): void {
    if (isNullOrUndefined(LocalStorageHelper.getItem(LocalStoragePartialEnum.personalizeCookie))) {
      this.openBanner();
    }
  }

  openBanner(): void {
    this._store.dispatch(new SolidifyAppAction.ChangeDisplayBannerCookieConsent(true));
  }

  hideBanner(): void {
    this._store.dispatch(new SolidifyAppAction.ChangeDisplayBannerCookieConsent(false));
  }

  notifyFeatureDisabledBecauseCookieDeclined(type: CookieType, name: string): void {
    const cookiePreference = this.listCookiePreferences.find(p => p.type === type && p.name === name);
    const label = this._translate.instant(cookiePreference.labelToTranslate);
    this._notificationService.showInformation(this.labelTranslate.notificationFeatureDisabledBecauseCookieDeclined
      , {label: label}, {
        text: this.labelTranslate.cookieConsentAllowTheCookie,
        callback: () => {
          this.changeCookieSetting(cookiePreference, true);
          this._notificationService.showSuccess(this.labelTranslate.notificationCookieXEnabled, {label: label});
        },
      });
  }
}
