/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Frontend - standard-errors-handler.service.ts
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

import {
  HttpErrorResponse,
  HttpStatusCode,
} from "@angular/common/http";
import {
  Injectable,
  Injector,
} from "@angular/core";
import {Store} from "@ngxs/store";
import {DefaultSolidifyEnvironment} from "../../environments/environment.solidify-defaults";
import {SolidifyError} from "../../errors/solidify.error";
import {ErrorHelper} from "../../helpers/error.helper";
import {ENVIRONMENT} from "../../injection-tokens/environment.injection-token";
import {LABEL_TRANSLATE} from "../../injection-tokens/label-to-translate.injection-token";
import {SolidifyHttpErrorResponseModel} from "../../models/errors/solidify-http-error-response.model";
import {LabelTranslateInterface} from "../../models/label-translate.model";
import {SolidifyAppAction} from "../../stores/abstract/app/app.action";
import {
  isInstanceOf,
  isNullOrUndefined,
} from "../../tools/is/is.tool";
import {ErrorsSkipperService} from "../errors-skipper.service";
import {NotificationService} from "../notification.service";
import {GlobalErrorsHandlerService} from "./global-errors-handler.service";

@Injectable()
export class StandardErrorsHandlerService extends GlobalErrorsHandlerService {
  // Error handling is important and needs to be loaded first.
  // Because of this we should manually inject the services with Injector.

  constructor(protected readonly _injector: Injector) {
    super(_injector);
  }

  handleError(error: Error | SolidifyHttpErrorResponseModel | SolidifyError<Error> | SolidifyError<SolidifyHttpErrorResponseModel>): void {
    const environment: DefaultSolidifyEnvironment = this._injector.get(ENVIRONMENT);
    const labelTranslate: LabelTranslateInterface = this._injector.get(LABEL_TRANSLATE);
    const errorSkipper: ErrorsSkipperService = this._injector.get(ErrorsSkipperService);
    const notificationService: NotificationService = this._injector.get(NotificationService);
    const nativeError = isInstanceOf(error, SolidifyError) ? error.nativeError : error as Error;
    if (nativeError === environment.errorToSkipInErrorHandler) {
      return;
    }
    const httpErrorKeyToSkipInErrorHandler = environment.httpErrorKeyToSkipInErrorHandler;

    if (nativeError instanceof HttpErrorResponse) {
      if (errorSkipper.shouldBeSkipped(nativeError.status)) {
        return;
      }
      if (nativeError.status === 0) {
        notificationService.showError(labelTranslate.notificationBackendUnavailable);
      }
      if (ErrorHelper.isErrorToTreat(nativeError, httpErrorKeyToSkipInErrorHandler)) {
        // Server Error
        if (nativeError.status === HttpStatusCode.Unauthorized) {
          if (isNullOrUndefined(nativeError.error) || nativeError.error.error === ErrorHelper.INVALID_TOKEN_ERROR) {
            const store = this._injector.get(Store);
            store.dispatch(new SolidifyAppAction.Logout());
          }
        }
      }
    }

    super.handleError(error);
  }
}
