/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Frontend - app.state.ts
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

import {
  HttpClient,
  HttpHeaders,
} from "@angular/common/http";
import {
  Optional,
  Type,
} from "@angular/core";
import {MatDialog} from "@angular/material/dialog";
import {TranslateService} from "@ngx-translate/core";
import {Navigate} from "@ngxs/router-plugin";
import {
  Action,
  Actions,
  ofActionErrored,
  StateContext,
  Store,
} from "@ngxs/store";
import * as moment from "moment";
import {HighlightLoader} from "ngx-highlightjs";
import {Observable} from "rxjs";
import {
  catchError,
  map,
  tap,
} from "rxjs/operators";
import {InfiniteLoginService} from "../../../auth/infinite-login.service";
import {OauthHelper} from "../../../auth/oauth.helper";
import {OAuth2Service} from "../../../auth/oauth2.service";
import {AbstractDialogInterface} from "../../../components/dialogs/abstract/abstract-dialog-interface.model";
import {AppConfigService} from "../../../config/app-config.service";
import {SOLIDIFY_CONSTANTS} from "../../../constants";
import {CookieType} from "../../../cookie-consent/enums/cookie-type.enum";
import {CookieConsentUtil} from "../../../cookie-consent/utils/cookie-consent.util";
import {ApiActionNamePartialEnum} from "../../../enums/partial/api-action-name-partial.enum";
import {LocalStoragePartialEnum} from "../../../enums/partial/local-storage-partial.enum";
import {StatePartialEnum} from "../../../enums/partial/state-partial.enum";
import {DefaultSolidifyEnvironment} from "../../../environments/environment.solidify-defaults";
import {JwtTokenHelper} from "../../../helpers/jwt-token.helper";
import {LocalStorageHelper} from "../../../helpers/local-storage.helper";
import {ApiService} from "../../../http/api.service";
import {HeaderEnum} from "../../../http/header.enum";
import {BackendModuleVersion} from "../../../models/backend-modules/backend-module-version.model";
import {SolidifyHttpErrorResponseModel} from "../../../models/errors/solidify-http-error-response.model";
import {FrontendVersion} from "../../../models/frontend-version.model";
import {GlobalBanner} from "../../../models/global-banner.model";
import {ActionSubActionCompletionsWrapper} from "../../../models/stores/base.action";
import {Token} from "../../../models/token.model";
import {NativeNotificationService} from "../../../services/native-notification.service";
import {NotificationService} from "../../../services/notification.service";
import {SolidifySecurityService} from "../../../services/security.service";
import {
  isArray,
  isFalse,
  isNotNullNorUndefined,
  isNullOrUndefined,
  isNullOrUndefinedOrWhiteString,
  isTrue,
} from "../../../tools/is/is.tool";
import {MappingObject} from "../../../types/mapping-type.type";
import {CacheBustingUtil} from "../../../utils/cache-busting.util";
import {DialogUtil} from "../../../utils/dialog.util";
import {MappingObjectUtil} from "../../../utils/mapping-object.util";
import {ofSolidifyActionCompleted} from "../../../utils/stores/store.tool";
import {StoreUtil} from "../../../utils/stores/store.util";
import {UserPreferencesUtil} from "../../../utils/user-preferences.util";
import {defaultBaseStateInitValue} from "../base/base.state";
import {BasicState} from "../base/basic.state";
import {SolidifyAppOptions} from "./app-options.model";
import {SolidifyAppStateModel} from "./app-state.model";
import {SolidifyAppAction} from "./app.action";

export const defaultSolidifyAppStateInitValue: () => SolidifyAppStateModel = () =>
  ({
    ...defaultBaseStateInitValue(),
    isTouchInterface: false,
    isApplicationInitialized: false,
    isServerOffline: false,
    isUnableToLoadApp: false,
    isLoggedIn: false,
    isInTourMode: false,
    ignorePreventLeavePopup: false,
    isOpenedSidebarUserGuide: false,
    isOpenedSidebarCookieConsent: false,
    isOpenedBannerCookieConsent: false,
    appLanguage: undefined,
    theme: undefined,
    darkMode: false,
    preventExit: false,
    preventExitReasonToTranslate: undefined,
    [StatePartialEnum.application_user]: undefined,
    token: undefined,
    frontendVersion: undefined,
    backendModulesVersion: {} as MappingObject<string, BackendModuleVersion>,
    backendModulesUrl: {} as MappingObject<string, string | string[]>,
    globalBanner: undefined,
  });

// @dynamic
export abstract class SolidifyAppState<TStateModel extends SolidifyAppStateModel> extends BasicState<TStateModel> {
  private static readonly _AUTHORIZATION_MODULE_NAME: string = "authorization";

  constructor(protected readonly _store: Store,
              protected readonly _translate: TranslateService,
              protected readonly _oauthService: OAuth2Service,
              protected readonly _actions$: Actions,
              protected readonly _apiService: ApiService,
              protected readonly _httpClient: HttpClient,
              protected readonly _notificationService: NotificationService,
              protected readonly _optionsState: SolidifyAppOptions,
              protected readonly _environment: DefaultSolidifyEnvironment,
              @Optional() protected readonly _highlightLoader: HighlightLoader,
              protected readonly _nativeNotificationService: NativeNotificationService,
              @Optional() protected readonly _securityService: SolidifySecurityService,
              protected readonly _appConfigService: AppConfigService,
              @Optional() protected readonly _dialog: MatDialog,
              @Optional() protected readonly _privacyPolicyTermsOfUseApprovalDialogType: Type<AbstractDialogInterface<any, any>>,
              @Optional() protected readonly _infiniteLoginService: InfiniteLoginService) {
    super();
    // _environment.appState = _options.baseAppSate;
    // _environment.appNameSpace = _options.nameSpace as SolidifyAppNameSpace;
  }

  loadModule(): ActionSubActionCompletionsWrapper[] | [] {
    return this._optionsState.loadModule ? [{
        action: new SolidifyAppAction.LoadModuleUrl(),
        subActionCompletions: [
          this._actions$.pipe(ofSolidifyActionCompleted(SolidifyAppAction.LoadModuleUrlSuccess)),
          this._actions$.pipe(ofSolidifyActionCompleted(SolidifyAppAction.LoadModuleUrlFail)),
        ],
      }]
      :
      [];
  }

  @Action(SolidifyAppAction.InitApplication)
  initApplication(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.InitApplication): Observable<boolean> {
    return StoreUtil.dispatchSequentialActionAndWaitForSubActionsCompletion(
      ctx,
      [
        {
          action: new SolidifyAppAction.ChangeAppLanguage(UserPreferencesUtil.getPreferredLanguage(this._environment)),
          subActionCompletions: [
            this._actions$.pipe(ofSolidifyActionCompleted(SolidifyAppAction.ChangeAppLanguageSuccess)),
          ],
        },
        ...this.loadModule(),
        {
          action: new SolidifyAppAction.Login(),
          subActionCompletions: [
            this._actions$.pipe(ofSolidifyActionCompleted(SolidifyAppAction.LoginSuccess)),
            this._actions$.pipe(ofActionErrored(SolidifyAppAction.LoginSuccess)),
            this._actions$.pipe(ofSolidifyActionCompleted(SolidifyAppAction.LoginFail)),
          ],
        },
        {
          action: new SolidifyAppAction.InitApplicationParallel(),
          subActionCompletions: [
            this._actions$.pipe(ofSolidifyActionCompleted(SolidifyAppAction.InitApplicationParallelSuccess)),
            this._actions$.pipe(ofSolidifyActionCompleted(SolidifyAppAction.InitApplicationParallelFail)),
          ],
        },
      ],
    ).pipe(
      map(result => {
        if (result.success) {
          ctx.dispatch(new SolidifyAppAction.InitApplicationSuccess(action, result));
        } else {
          ctx.dispatch(new SolidifyAppAction.InitApplicationFail(action, result));
        }
        return result.success;
      }),
    );
  }

  @Action(SolidifyAppAction.InitApplicationParallel)
  initApplicationParallel(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.InitApplicationParallel): Observable<boolean> {
    return StoreUtil.dispatchParallelActionAndWaitForSubActionsCompletion(
      ctx,
      [
        {action: new SolidifyAppAction.DefineFallbackLanguage()},
        {action: new SolidifyAppAction.SetMomentLocal()},
        ...this._optionsState.initApplicationParallel,
      ],
    ).pipe(
      map(result => {
        if (result.success) {
          ctx.dispatch(new SolidifyAppAction.InitApplicationParallelSuccess(action));
        } else {
          ctx.dispatch(new SolidifyAppAction.InitApplicationParallelFail(action));
        }
        return result.success;
      }),
    );
  }

  @Action(SolidifyAppAction.LoadModuleUrl)
  loadModuleUrl(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.LoadModuleUrl): Observable<MappingObject<string, string>> {
    if (isNullOrUndefined(this._environment.allowedUrls)) {
      this._environment.allowedUrls = [];
    }
    const url = this._optionsState.urlAdminModule();
    this._environment.allowedUrls.push(url);
    return this._apiService.get<MappingObject<string, string>>(url)
      .pipe(
        tap(runtimeConfig => {
          // throw new Error("SIMULATE SERVER OFFLINE");
          ctx.dispatch(new SolidifyAppAction.LoadModuleUrlSuccess(action, runtimeConfig));
        }),
        catchError((error: SolidifyHttpErrorResponseModel) => {
          ctx.dispatch(new SolidifyAppAction.LoadModuleUrlFail(action));
          throw error;
        }),
      );
  }

  @Action(SolidifyAppAction.LoadModuleUrlSuccess)
  loadModuleUrlSuccess(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.LoadModuleUrlSuccess): void {
    const initialAuthorizationUrl = this._environment.authorization;
    Object.assign(this._environment, action.modulesUrl);
    ctx.patchState({
      backendModulesUrl: action.modulesUrl,
    });
    if (isFalse(this._environment.computeOAuthEndpoint)) {
      this._environment.authorization = initialAuthorizationUrl;
    }
    if (isNullOrUndefinedOrWhiteString(this._environment.authorization)) {
      // eslint-disable-next-line no-console
      console.warn("'authorization' property should not be empty. Please set the value on backend side or disable 'computeOAuthEndpoint' and set 'authorization' on portal environment file");
    }
    const tokenEndpoint = OauthHelper.getTokenEndpoint(this._environment.authorization);
    const loginUrl = OauthHelper.getLoginUrl(this._environment.authorization);
    this._environment.allowedUrls.push(tokenEndpoint, loginUrl);
  }

  @Action(SolidifyAppAction.LoadModuleUrlFail)
  loadModuleUrlFail(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.LoadModuleUrlFail): void {
    ctx.patchState({
      isServerOffline: true,
    });
    if (isFalse(this._environment.maintenanceMode)) {
      ctx.dispatch(new Navigate([this._optionsState.routeServerOffline]));
    }
  }

  @Action(SolidifyAppAction.LoadBackendModulesVersion)
  loadBackendModulesVersion(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.LoadBackendModulesVersion): Observable<boolean> {
    const actionSubActionCompletionsWrapper: ActionSubActionCompletionsWrapper[] = [];
    MappingObjectUtil.forEach(action.modulesUrl, (value, key) => {
      if (isNullOrUndefinedOrWhiteString(value)) {
        return;
      }
      if (isArray(value)) {
        value.forEach((v, index) => {
          actionSubActionCompletionsWrapper.push({
            action: new SolidifyAppAction.LoadBackendModuleVersion(key + "-" + (index + 1), v),
            subActionCompletions: [
              this._actions$.pipe(ofSolidifyActionCompleted(SolidifyAppAction.LoadBackendModuleVersionSuccess)),
              this._actions$.pipe(ofSolidifyActionCompleted(SolidifyAppAction.LoadBackendModuleVersionFail)),
            ],
          });
        });
      } else {
        actionSubActionCompletionsWrapper.push({
          action: new SolidifyAppAction.LoadBackendModuleVersion(key, value),
          subActionCompletions: [
            this._actions$.pipe(ofSolidifyActionCompleted(SolidifyAppAction.LoadBackendModuleVersionSuccess)),
            this._actions$.pipe(ofSolidifyActionCompleted(SolidifyAppAction.LoadBackendModuleVersionFail)),
          ],
        });
      }
    });

    return StoreUtil.dispatchParallelActionAndWaitForSubActionsCompletion(this._store, actionSubActionCompletionsWrapper).pipe(
      map(result => {
        if (isFalse(result.success)) {
          ctx.dispatch(new SolidifyAppAction.LoadBackendModulesVersionFail(action));
          return false;
        }
        ctx.dispatch(new SolidifyAppAction.LoadBackendModulesVersionSuccess(action, ctx.getState().backendModulesVersion));
        return true;
      }),
    );
  }

  @Action(SolidifyAppAction.LoadBackendModulesVersionSuccess)
  loadBackendModulesVersionSuccess(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.LoadBackendModulesVersionSuccess): void {
    const modulesVersion = MappingObjectUtil.copy(ctx.getState().backendModulesVersion);
    ctx.patchState({
      backendModulesVersion: modulesVersion,
    });
  }

  @Action(SolidifyAppAction.LoadBackendModulesVersionFail)
  loadBackendModulesVersionFail(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.LoadBackendModulesVersionFail): void {
  }

  @Action(SolidifyAppAction.LoadBackendModuleVersion)
  loadBackendModuleVersion(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.LoadBackendModuleVersion): Observable<BackendModuleVersion> {
    const url = SolidifyAppState._getModuleInfoUrl(action.moduleName, action.moduleUrl);
    let headers = new HttpHeaders();
    headers = headers.set(HeaderEnum.skipOauth2TokenHeader, SOLIDIFY_CONSTANTS.STRING_TRUE);

    return this._httpClient.get<BackendModuleVersion>(url, {
      headers: headers,
    })
      .pipe(
        tap(backendModuleVersion => {
          backendModuleVersion.urlInfo = url;
          ctx.dispatch(new SolidifyAppAction.LoadBackendModuleVersionSuccess(action, backendModuleVersion));
        }),
        catchError((error: SolidifyHttpErrorResponseModel) => {
          ctx.dispatch(new SolidifyAppAction.LoadBackendModuleVersionFail(action));
          throw this._environment.errorToSkipInErrorHandler;
        }),
      );
  }

  private static _getModuleInfoUrl(name: string, url: string): string {
    if (name !== this._AUTHORIZATION_MODULE_NAME && (url.match(/\//g) || []).length > 2) {
      url = url.substring(0, url.lastIndexOf(SOLIDIFY_CONSTANTS.URL_SEPARATOR));
    }
    return url + SOLIDIFY_CONSTANTS.URL_SEPARATOR + ApiActionNamePartialEnum.INFO;
  }

  @Action(SolidifyAppAction.LoadBackendModuleVersionSuccess)
  loadBackendModuleVersionSuccess(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.LoadBackendModuleVersionSuccess): void {
    const mappingBackendModulesVersion = MappingObjectUtil.copy(ctx.getState().backendModulesVersion);
    const backendModuleVersion = action.backendModuleVersion;
    backendModuleVersion.name = action.parentAction.moduleName;
    backendModuleVersion.url = action.parentAction.moduleUrl;
    MappingObjectUtil.set(mappingBackendModulesVersion, action.parentAction.moduleName, backendModuleVersion);
    ctx.patchState({
      backendModulesVersion: mappingBackendModulesVersion,
    });
  }

  @Action(SolidifyAppAction.LoadBackendModuleVersionFail)
  loadBackendModuleVersionFail(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.LoadBackendModuleVersionFail): void {
  }

  @Action(SolidifyAppAction.InitApplicationSuccess)
  initApplicationSuccess(ctx: StateContext<SolidifyAppStateModel>): void {
    ctx.patchState({
      isApplicationInitialized: true,
    });
    this.subscribe(this._appConfigService.environmentUpdatedObs.pipe(
      tap(environment => {
        this._navigateToMaintenanceIfApplicable();
      }),
    ));
    this._navigateToMaintenanceIfApplicable();
  }

  private _navigateToMaintenanceIfApplicable(): boolean {
    if (isTrue(this._environment.maintenanceMode)
      && (isNullOrUndefined(this._securityService) || isFalse(this._securityService.isLoggedIn()) || isFalse(this._securityService.isRoot()))) {
      this._store.dispatch(new Navigate([this._optionsState.routeMaintenance]));
      return true;
    }
    return false;
  }

  @Action(SolidifyAppAction.InitApplicationFail)
  initApplicationFail(ctx: StateContext<SolidifyAppStateModel>): void {
    if (this._navigateToMaintenanceIfApplicable()) {
      return;
    }
    if (isFalse(ctx.getState().isServerOffline)) {
      ctx.patchState({
        isUnableToLoadApp: true,
      });
      ctx.dispatch(new Navigate([this._optionsState.routeUnableToLoadApp]));
    }
  }

  @Action(SolidifyAppAction.DefineFallbackLanguage)
  defineFallbackLanguage(ctx: StateContext<SolidifyAppStateModel>): void {
    // this language will be used as a fallback when a translation isn't found in the current language
    this._translate.setDefaultLang(this._environment.defaultLanguage);
  }

  @Action(SolidifyAppAction.SetMomentLocal)
  setMomentLocal(ctx: StateContext<SolidifyAppStateModel>): void {
    const navigatorLanguage = navigator.language;
    moment.locale(navigatorLanguage);
    // this._adapter.setLocale("fr-CH"); // TOFIX : DOESN'T WORK HERE, NEED TO BE SET IN APP MODULE
  }

  @Action(SolidifyAppAction.ChangeAppLanguage)
  changeAppLanguage(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.ChangeAppLanguage): Observable<any> {
    ctx.patchState({
      appLanguage: action.language,
    });

    if (CookieConsentUtil.isFeatureAuthorized(CookieType.localStorage, LocalStoragePartialEnum.language)) {
      LocalStorageHelper.setItem(LocalStoragePartialEnum.language, action.language);
    }
    return this._translate.use(action.language).pipe(
      tap(() => {
        ctx.dispatch(new SolidifyAppAction.ChangeAppLanguageSuccess(action));
      }),
    );
  }

  @Action(SolidifyAppAction.ChangeAppLanguageSuccess)
  changeAppLanguageSuccess(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.ChangeAppLanguageSuccess): void {
    // ctx.dispatch(new AppCarouselAction.GetCarousel());
  }

  @Action(SolidifyAppAction.ChangeAppTheme)
  changeAppTheme(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.ChangeAppTheme): void {
    this._environment.theme = action.theme;
    ctx.patchState({
      theme: action.theme,
    });
    // Force to refresh language file consolidated with theme json and customizable json
    this._environment.appLanguages.forEach(language => this._translate.reloadLang(language));
    // Force to refresh current page translation by switching of language (only possible if more than one language)
    const currentLanguage = ctx.getState().appLanguage;
    const isFirstLoadApp = isNullOrUndefinedOrWhiteString(currentLanguage);
    if (this._environment.appLanguages.length > 1 && !isFirstLoadApp) {
      const otherLanguage = this._environment.appLanguages.find(l => l !== currentLanguage);
      this.subscribe(StoreUtil.dispatchActionAndWaitForSubActionCompletion(this._store, this._actions$,
        new SolidifyAppAction.ChangeAppLanguage(otherLanguage),
        SolidifyAppAction.ChangeAppLanguageSuccess,
        () => {
          ctx.dispatch(new SolidifyAppAction.ChangeAppLanguage(currentLanguage));
        }));
    }
  }

  @Action(SolidifyAppAction.ChangeDarkMode)
  changeDarkMode(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.ChangeDarkMode): void {
    ctx.patchState({
      darkMode: action.enabled,
    });
    if (isNotNullNorUndefined(this._highlightLoader)) {
      this._highlightLoader.setTheme(action.enabled ? this._environment.highlightJsThemeDark : this._environment.highlightJsThemeLight);
    }
    if (CookieConsentUtil.isFeatureAuthorized(CookieType.localStorage, LocalStoragePartialEnum.darkMode)) {
      LocalStorageHelper.setItem(LocalStoragePartialEnum.darkMode, action.enabled ? SOLIDIFY_CONSTANTS.STRING_TRUE : SOLIDIFY_CONSTANTS.STRING_FALSE);
    }
  }

  @Action(SolidifyAppAction.ChangeDisplaySidebarUserGuide)
  changeDisplaySidebarUserGuide(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.ChangeDisplaySidebarUserGuide): void {
    ctx.patchState({
      isOpenedSidebarUserGuide: action.isOpen,
    });
  }

  @Action(SolidifyAppAction.ChangeDisplayBannerCookieConsent)
  displayCookieConsent(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.ChangeDisplayBannerCookieConsent): void {
    ctx.patchState({
      isOpenedBannerCookieConsent: action.isOpen,
    });
  }

  @Action(SolidifyAppAction.ChangeDisplaySidebarCookieConsent)
  changeDisplaySidebarCookieConsent(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.ChangeDisplaySidebarCookieConsent): void {
    ctx.patchState({
      isOpenedSidebarCookieConsent: action.isOpen,
    });
  }

  @Action(SolidifyAppAction.Login)
  login(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.Login): Observable<boolean> {
    return this._oauthService.tryLogin()
      .pipe(
        tap(e => {
            if (e) {
              ctx.dispatch(new SolidifyAppAction.LoginSuccess(action));
            } else {
              ctx.dispatch(new SolidifyAppAction.LoginFail(action));
            }
          },
        ));
  }

  @Action(SolidifyAppAction.LoginSuccess)
  loginSuccess(ctx: StateContext<SolidifyAppStateModel>): Token {
    const accessToken = this._oauthService.getAccessToken();
    const tokenDecoded = JwtTokenHelper.decodeToken(accessToken) as Token;
    ctx.patchState({
      isLoggedIn: true,
      token: tokenDecoded,
    });
    this.subscribe(this._nativeNotificationService.requestNotification());
    if (isNotNullNorUndefined(this._infiniteLoginService)) {
      this._infiniteLoginService.resetRemainingTryLogging();
    }
    return tokenDecoded;
  }

  @Action(SolidifyAppAction.LoginFail)
  loginFail(ctx: StateContext<SolidifyAppStateModel>): void {
    if (isNotNullNorUndefined(this._infiniteLoginService)) {
      this._infiniteLoginService.decrementTryLogging();
    }
    if (isTrue(this._optionsState.forceLogin) && !ctx.getState().isServerOffline && !ctx.getState().isUnableToLoadApp) {
      ctx.dispatch(new SolidifyAppAction.InitiateNewLogin());
    } else {
      ctx.patchState({
        isLoggedIn: false,
      });
    }
  }

  @Action(SolidifyAppAction.GetAppVersion)
  getAppVersion(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.GetAppVersion): Observable<FrontendVersion> {
    return this._httpClient.get<FrontendVersion>("./assets/" + this._environment.frontendVersionFile + CacheBustingUtil.generateCacheBustingQueryParam())
      .pipe(
        tap(result => {
          ctx.dispatch(new SolidifyAppAction.GetAppVersionSuccess(action, result));
        }),
        catchError((error: SolidifyHttpErrorResponseModel) => {
          ctx.dispatch(new SolidifyAppAction.GetAppVersionFail(action));
          throw error;
        }),
      );
  }

  @Action(SolidifyAppAction.GetAppVersionSuccess)
  getAppVersionSuccess(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.GetAppVersionSuccess): void {
    ctx.patchState({
      frontendVersion: action.appVersion,
    });
  }

  @Action(SolidifyAppAction.GetAppVersionFail)
  getAppVersionFail(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.GetAppVersionFail): void {
  }

  @Action(SolidifyAppAction.Logout)
  logout(ctx: StateContext<SolidifyAppStateModel>): void {
    const isLoggedIn = ctx.getState().isLoggedIn;
    if (isFalse(isLoggedIn)) {
      this._oauthService.logOut(true);
      return; // PREVENT INFINITE CANCEL LOOP DUE TO MULTIPLE CALL TO THIS METHOD THAT REDIRECT TO HOME PAGE
    }
    ctx.patchState({
      ignorePreventLeavePopup: true,
      isLoggedIn: false,
      token: undefined,
    });

    // Revoke tokens servers side
    this.subscribe(this._apiService.post(this._optionsState.urlRevokeToken()));
    setTimeout(() => {
      this._treatmentAfterRevokeToken(ctx);
    }, 200);
  }

  protected _treatmentAfterRevokeToken(ctx: StateContext<SolidifyAppStateModel>): void {
    ctx.dispatch(new Navigate([this._environment.routeHomePage]));
    this._oauthService.logOut(true);

    ctx.patchState({
      ignorePreventLeavePopup: false,
    });
    if (isTrue(this._environment.reloadAfterLogout)) {
      window.location.reload();
    } else {
      this._navigateToMaintenanceIfApplicable();
    }
  }

  @Action(SolidifyAppAction.CancelPreventExit)
  cancelPreventExit(ctx: StateContext<SolidifyAppStateModel>): void {
    ctx.patchState({
      preventExit: false,
      preventExitReasonToTranslate: undefined,
    });
  }

  @Action(SolidifyAppAction.PreventExit)
  preventExit(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.PreventExit): void {
    ctx.patchState({
      preventExit: true,
      preventExitReasonToTranslate: action.reasonToTranslate,
    });
  }

  @Action(SolidifyAppAction.SetIsInTourMode)
  setIsInTourMode(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.SetIsInTourMode): void {
    ctx.patchState({
      isInTourMode: action.isInTourMode,
    });
  }

  @Action(SolidifyAppAction.SetIsTouchInterface)
  setIsTouchInterface(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.SetIsTouchInterface): void {
    ctx.patchState({
      isTouchInterface: action.isTouchInterface,
    });
  }

  @Action(SolidifyAppAction.GetGlobalBanner)
  getGlobalBanner(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.GetGlobalBanner): Observable<any> {
    if (isNullOrUndefinedOrWhiteString(this._optionsState.urlActiveGlobalBanner)) {
      return;
    }
    let headers = new HttpHeaders();
    headers = headers.set(HeaderEnum.skipOauth2TokenHeader, SOLIDIFY_CONSTANTS.STRING_TRUE);

    return this._httpClient.get<GlobalBanner>(this._optionsState.urlActiveGlobalBanner(), {
      headers: headers,
    })
      .pipe(
        tap(result => {
          ctx.dispatch(new SolidifyAppAction.GetGlobalBannerSuccess(action, result));
        }),
        catchError((error: SolidifyHttpErrorResponseModel) => {
          ctx.dispatch(new SolidifyAppAction.GetGlobalBannerFail(action));
          throw this._environment.errorToSkipInErrorHandler;
        }),
      );
  }

  @Action(SolidifyAppAction.GetGlobalBannerSuccess)
  getGlobalBannerSuccess(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.GetGlobalBannerSuccess): void {
    ctx.patchState({
      globalBanner: action.globalBanner,
    });
  }

  @Action(SolidifyAppAction.GetGlobalBannerFail)
  getGlobalBannerFail(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.GetGlobalBannerFail): void {
    ctx.patchState({
      globalBanner: undefined,
    });
  }

  @Action(SolidifyAppAction.InitiateNewLogin)
  initiateNewLogin(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.InitiateNewLogin): void {
    const privacyPolicyAndTermsOfUseAccepted = LocalStorageHelper.getItem(LocalStoragePartialEnum.privacyPolicyAndTermsOfUseAccepted);
    if (isNotNullNorUndefined(this._dialog)
      && isNotNullNorUndefined(this._privacyPolicyTermsOfUseApprovalDialogType)
      && isTrue(this._environment.displayPrivacyPolicyAndTermsOfUseApprovalDialog)
      && (isNotNullNorUndefined(this._environment.privacyPolicyLink) || isNotNullNorUndefined(this._environment.termsOfUseLink))
      && privacyPolicyAndTermsOfUseAccepted !== this._environment.versionPrivacyPolicyAndTermsOfUse) {
      this.subscribe(DialogUtil.open(this._dialog, this._privacyPolicyTermsOfUseApprovalDialogType, {
        termsOfUseLink: this._environment.termsOfUseLink,
        privacyPolicyLink: this._environment.privacyPolicyLink,
        language: ctx.getState().appLanguage,
      }, {
        takeOne: true,
      }, result => {
        if (isTrue(result)) {
          if (CookieConsentUtil.isFeatureAuthorized(CookieType.localStorage, LocalStoragePartialEnum.privacyPolicyAndTermsOfUseAccepted)) {
            LocalStorageHelper.setItem(LocalStoragePartialEnum.privacyPolicyAndTermsOfUseAccepted, this._environment.versionPrivacyPolicyAndTermsOfUse);
          }
          this._oauthService.initAuthorizationCodeFlow();
        }
      }));
    } else {
      this._oauthService.initAuthorizationCodeFlow();
    }
  }
}
